# !/bin/bash
#
# Daemon waits for an Update Trigger, and runs the update process when it gets one.
# Copyright (c) RAYLASE GmbH
#

VERSION="V1.0"

export PATH="${PATH:+$PATH:}/usr/sbin:/sbin:/opt/RAYLASE/scripts"

# Variable configuration and function definitions
. upd_functions.sh

# Error code definitions
. upd_err.sh

SFC_DIR="/opt/RAYLASE/WebInterface"
UPDATEPROGRESS_FILE="${SFC_DIR}/Content/update.txt"
UPDATEONGIONG_FILE="/tmp/updateOngoing"

#
# Construct an error message from the symbolic errorcode name,
# then tack on any other supplied args.
# Send the message to the update log file, and to the
# update progress text file, which the WebIF reads periodically.
#
output_err_msg() {
	local errname="$1"

	# insert the numeric value of the errorcode, followed by the associated error text.
	local errmsg="${!errname}: ${UPD_ERR_MSG[${errname}]}"

	# see if there is anything else to add to the message.
	if (($# > 1)) ; then
		shift
		errmsg="${errmsg}: $*"
	fi
	logecho "Error: ${errmsg}"
	echo "${errmsg}" > "$UPDATEPROGRESS_FILE"
}

init_daemon() {
	if [ -f ${LOG_FILE} ] ; then
		cat /dev/null > "${LOG_FILE}"
	fi

	logecho "Initialising Update Daemon $VERSION for SFC."

	if [ "$(whoami)" != "root" ] ; then
		fatal "Must be root."
	fi

	# Check & log which image is running currently
	check_booted_image

	DAEMON_PID=$$
	logecho "PID of Update Daemon = ${DAEMON_PID}"
}

# $1 tar package name
# $2 destination directory
# $3 physical partition number (1..7)
# $4 logical partition number (0..2)
install_partition() {

	echo 15 > "$UPDATEPROGRESS_FILE"

	if [ -z "$4" ] ; then
		output_err_msg 'UPD_ERR_NOIMGNUM'
		return 1
	fi
	if [ -z "$3" ] ; then
		output_err_msg 'UPD_ERR_NOIMGPART'
		return 1
	fi
	if [ -z "$2" ] ; then
		output_err_msg 'UPD_ERR_NODEST'
		return 1
	fi
	if [ -z "$1" ] ; then
		output_err_msg 'UPD_ERR_NOPKGNAME'
		return 1
	fi
	
	if [ ! -f "$1" ] ; then
		output_err_msg 'UPD_ERR_PKGNOTEXIST' "$1"
		return 1
	fi

	local updpkg="$1"
	local DEV_ID="$3"
	local MOUNT_DIR="$2"
	local PART_LABEL="$4"
		
	umount -q "${MOUNT_DIR}"

	logecho "Formatting partition ${DISKDEVICE}${DEV_ID} as ${PART_LABEL}"
	mkfs.ext4 -q -F -L "${PART_LABEL}" "${DISKDEVICE}${DEV_ID}"
	
	logecho "Mounting ${DISKDEVICE}${DEV_ID} on ${MOUNT_DIR}"
	mount "${DISKDEVICE}${DEV_ID}" "${MOUNT_DIR}"
	echo 25 > $UPDATEPROGRESS_FILE
	
	logecho "Extracting contents of '$1' to ${MOUNT_DIR}"
	tar -xf "${updpkg}" -C "${MOUNT_DIR}" >> "${LOG_FILE}" 2>&1
	
	if [ $? != 0 ] ; then
		output_err_msg 'UPD_ERR_PKGEXTRFAIL' "${updpkg}"
		return 1
	fi
	
	cp ${MOUNT_DIR}/boot/bzImage /boot/efi/bzImage-${PART_LABEL}
	
	if [ ${DEV_ID} -ne ${IMG0_DEVID} ] ; then
		restore_etc
	fi
	
	echo 60 > $UPDATEPROGRESS_FILE
	logecho "Extraction finished successfully"
	return 0;
}

delete_used_package() {
	if [ -f ${updpkg} ] ; then
		logecho "Deleting used package '${updpkg}'"
		rm -f ${updpkg}
	fi
}

# Perform update
do_update() {
	logecho "Update procedure started: using '${updpkg}'"
	echo 10 > $UPDATEPROGRESS_FILE

	local updpkg="$1"
	local MOUNT_DIR="$2"
	local DEV_ID="$3"
	local PART_LABEL="$4"

	install_partition "${updpkg}" "${MOUNT_DIR}" "${DEV_ID}" "${PART_LABEL}"
	local unpack_data_return_code=$?
	
	if [ ${unpack_data_return_code} == 0 ] ; then
		logecho "Unmounting ${DISKDEVICE}${DEV_ID} from ${MOUNT_DIR}"
		
		local unmounted="false"
		for (( i=0; i < 10; ++i )) ; do
			if sudo umount "${DISKDEVICE}${DEV_ID}" &&	sudo e2fsck -f -n "${DISKDEVICE}${DEV_ID}" ; then
				unmounted="true"
				break
			fi
			logecho "Retrying unmount $i"
			sleep 1
		done
		echo 66 > $UPDATEPROGRESS_FILE

		if [ "${unmounted}" == "true" ] ; then
			logecho "Unmounting ${MOUNT_DIR} succeeded"
			echo "68" > "$UPDATEPROGRESS_FILE"
		else
			output_err_msg 'UPD_ERR_UMNTIMGDIRFAIL' "'${MOUNT_DIR}'"
		fi
	fi
}

start_update() {
	echo "1" > "$UPDATEPROGRESS_FILE"
	logecho "SIGHUP received, looking for an update package"

	if [ -n "$(find ${UPD_DIR} -maxdepth 1 -name 'update-sfc*.tar.bz2' -print -quit)" ]; then
		echo "5" > "$UPDATEPROGRESS_FILE"
		
		touch ${UPDATEONGIONG_FILE}
		send_signal_to_firmware
		
		local PART_LABEL='1'
		local MOUNT_DIR="/mnt"
		local DEV_ID="${IMG1_DEVID}"
		
		if [ "${BOOTED_IMAGE}" == '0' ] ; then
			PART_ID='1'
			PART_LABEL='SYS1'
			DEV_ID="${IMG1_DEVID}"
		elif [ "${BOOTED_IMAGE}" == '1' ] ; then
			PART_ID='2'
			PART_LABEL='SYS2'
			DEV_ID="${IMG2_DEVID}"
		elif [ "${BOOTED_IMAGE}" == '2' ] ; then
			PART_ID='1'
			PART_LABEL='SYS1'
			DEV_ID="${IMG1_DEVID}"
		fi

		for updpkg in `ls ${UPD_DIR}/update-sfc*.tar.bz2` ; do
			(( ++foundPackages ))

			logecho "Unpack package: '${updpkg}'"
			bunzip2 ${updpkg}
			
			updpkg=$(echo ${updpkg} | sed 's/\.bz2//')
			logecho "Update package: '${updpkg}'"
			
			do_update "$updpkg" "$MOUNT_DIR" "$DEV_ID" "$PART_LABEL"

			logecho "Refreshing GRUB environment after update"
			write_grub_boot_device ${PART_ID}
			
			logecho "---- UPDATE FINISHED: REBOOTING NOW -----"
			rotate_logfile
			rm -f ${UPDATEONGIONG_FILE}

			reboot
		done
	fi
	
	if (( foundPackages < 1 )); then
		output_err_msg 'UPD_ERR_NOPKGINDIR' "'${UPD_DIR}'"
	fi
	sleep 1
}

check_recover() {
	if [ -n "$(find ${UPD_DIR} -maxdepth 1 -name 'update-sfc*.tar' -print -quit)" ]; then
		echo "5" > "$UPDATEPROGRESS_FILE"
		
		touch ${UPDATEONGIONG_FILE}
		send_signal_to_firmware

		local PART_LABEL='RECOVER'
		local MOUNT_DIR="/mnt"
		local DEV_ID="${IMG0_DEVID}"
			
		for updpkg in `ls ${UPD_DIR}/update-sfc*.tar` ; do
			logecho "Found update package: '${updpkg}'"

			if [ "${BOOTED_IMAGE}" != '0' ] ; then
				# Give some time to relax
				sync
				
				PARTITION=${DISKDEVICE}${IMG0_DEVID}

				VERSION=$(tar -xf ${updpkg} -O ./etc/recover.conf)
				VERSION=$(echo ${VERSION} | cut -d' ' -f2)
				logecho "Min sfc version of recover image \"${VERSION}\""
				MIN_VERSION=(${VERSION//./ })
				
				if [ -z ${VERSION} ] ; then
					logecho "Empty min version, delete ${updpkg} and return"
					delete_used_package
					return
				fi

				umount -q /mnt
				mount ${PARTITION} /mnt

				VERSION=$(grep VERSION /mnt/sfc-image-version | cut -d= -f2 | sed 's/\"//g' )
				logecho "Actual sfc version of recover partition \"${VERSION}\""
				REC_VERSION=(${VERSION//./ })

				umount /mnt

				if [ -z ${VERSION} ] ; then
					logecho "Empty sfc version in recover partition"
					INSTALL_REQUIRED=1
				elif [ ${#REC_VERSION[@]} -ne 4 ] ; then
					logecho "Wrong sfc version in recover partition"
					INSTALL_REQUIRED=1
				elif [ ${REC_VERSION[0]} -lt ${MIN_VERSION[0]} ] ; then
					logecho "First position of sfc version in recover partition less than required"
					INSTALL_REQUIRED=1
				elif [ ${REC_VERSION[0]} -eq ${MIN_VERSION[0]} ] ; then
					if [ ${REC_VERSION[1]} -lt ${MIN_VERSION[1]} ] ; then
						logecho "Second position of sfc version in recover partition less than required"
						INSTALL_REQUIRED=1
					elif [ ${REC_VERSION[1]} -eq ${MIN_VERSION[1]} ] ; then
						if [ ${REC_VERSION[2]} -lt ${MIN_VERSION[2]} ] ; then
							logecho "Third position of sfc version in recover partition less than required"
							INSTALL_REQUIRED=1
						elif [ ${REC_VERSION[2]} -eq ${MIN_VERSION[2]} ] ; then
							if [ ${REC_VERSION[3]} -lt ${MIN_VERSION[3]} ] ; then
								logecho "Fourth position of sfc version in recover partition less than required"
								INSTALL_REQUIRED=1
							fi
						fi
					fi
				fi

				if [ ${INSTALL_REQUIRED} -eq 1 ] ; then
					do_update "$updpkg" "$MOUNT_DIR" "$DEV_ID" "$PART_LABEL"
					delete_used_package
					logecho "---- UPDATE FINISHED RECOVER partion) -----"
					rotate_logfile

					rm -f ${UPDATEONGIONG_FILE}
					send_signal_to_firmware
				fi
			fi
			
			delete_used_package
		done
	fi
	sleep 1
}

# -----------------------------------------------------------------
#
# main function
#

init_daemon

logecho "Update Daemon $VERSION for SFC started."
echo "0" > "$UPDATEPROGRESS_FILE"
chmod 666 $UPDATEPROGRESS_FILE

check_recover

# Start an update, after receiving the signal
trap start_update SIGHUP

while :
do
	# Simply do nothing
	sleep 120 &
	wait $!
done
